"""
Analyze features.

Usage:
 ./classes/features.py

Author:
 Peter Rigali - 2022-03-19
"""
from dataclasses import dataclass
from typing import List, Optional
import numpy as np
import pandas as pd
import statsmodels.api as sm
from pyjr.classes.model_data import ModelingData
from pyjr.utils.base import _mean, _percentile, _variance, _sum, _std, _min
from pyjr.utils.tools import _add_constant, _round_to, _to_metatype, _cent, _dis, stack


@dataclass
class FeaturePerformance:

    __slots__ = ["modeling_data", "reg_results", "outlier_results"]

    def __init__(self, data: ModelingData):
        self.modeling_data = data
        self.reg_results = None
        self.outlier_results = None

    def __repr__(self):
        return "FeaturePerformance"

    def add_regression(self):
        """Return r2, pred to true correlation, and mean of residuals"""
        results = {}
        dic = dict(zip(range(len(self.modeling_data.x_data_names)), self.modeling_data.x_data_names))
        for i in range(len(self.modeling_data.x_data_names)):
            key = dic[i]
            results[key] = {}
            x = _add_constant(data=self.modeling_data.x_train[:, i])
            y = self.modeling_data.y_train
            lin_reg = sm.OLS(y, x).fit()
            pred = lin_reg.predict(_add_constant(data=self.modeling_data.x_test[:, i]))
            flat_ytest = self.modeling_data.y_test.reshape(1, pred.shape[0]).tolist()[0]
            results[key]["r2"] = _round_to(data=lin_reg.rsquared, val=100, remainder=True)
            results[key]['pred_true_coef'] = _round_to(data=np.corrcoef(pred, flat_ytest)[0, 1], val=100, remainder=True)
            results[key]['residuals_mean'] = _round_to(data=lin_reg.resid.mean(), val=100, remainder=True)
        self.reg_results = results
        return self

    def add_outlier_std(self, plus: bool = True, std_value: int = 2, return_ind: bool = True):
        per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
        dic = {val: self.modeling_data.x_data[:, ind] for ind, val in enumerate(self.modeling_data.x_data_names)}
        for key, val in dic.items():
            new_data = val.tolist()
            if _min(new_data) >= 0:
                if plus:
                    ind = np.where(val <= _percentile(data=new_data, q=per_dic[std_value]))[0]
                else:
                    ind = np.where(val >= _percentile(data=new_data, q=per_dic[-std_value]))[0]
            else:
                if plus:
                    ind = np.where(val <= _mean(data=new_data) + _std(data=new_data) * std_value)[0]
                else:
                    ind = np.where(val >= _mean(data=new_data) - _std(data=new_data) * std_value)[0]
            if return_ind:
                dic[key] = tuple(ind.tolist())
            else:
                dic[key] = tuple(val[ind].tolist())
        self.outlier_results = dic
        return self

    def add_outlier_var(self, plus: Optional[bool] = True, std_value: int = 2, return_ind: bool = True):
        per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
        dic = {val: self.modeling_data.x_data[:, ind] for ind, val in enumerate(self.modeling_data.x_data_names)}
        for key, val in dic.items():
            lst = val.tolist()
            temp_var = _variance(data=lst)
            dev_based = np.array([temp_var - _variance(np.delete(lst, i)) for i, j in enumerate(lst)])
            if plus:
                q = _percentile(data=lst, q=per_dic[std_value])
                ind = np.where(dev_based <= q)[0]
            else:
                q = _percentile(data=lst, q=per_dic[-std_value])
                ind = np.where(dev_based >= q)[0]

            if return_ind:
                dic[key] = tuple(ind.tolist())
            else:
                dic[key] = tuple(val[ind].tolist())
        self.outlier_results = dic
        return self

    def add_outlier_regression(self, plus: Optional[bool] = True, std_value: Optional[int] = 2, return_ind: bool = True):
        per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
        dic = {val: self.modeling_data.x_data[:, ind] for ind, val in enumerate(self.modeling_data.x_data_names)}
        for key, val in dic.items():
            arr = stack(val, self.modeling_data.y_data, False)
            ran = np.array(range(self.modeling_data.len))
            mu_y = np.zeros(len(arr) - 1)
            line_ys = []
            for i, j in enumerate(arr):
                xx, yy = np.delete(arr[:, 0], i), np.delete(arr[:, 1], i)
                w1 = (np.cov(xx, yy, ddof=1) / _variance(xx, ddof=1))[0, 1]
                new_y = w1 * ran[:-1] + (-1 * _mean(xx) * w1 + _mean(yy))
                mu_y = (mu_y + new_y) / 2
                line_ys.append(new_y)

            reg_based = np.array([np.mean(np.square(mu_y - j)) for i, j in enumerate(line_ys)])
            if plus:
                threshold = _percentile(data=reg_based, q=per_dic[std_value])
                ind = np.where(reg_based <= threshold)[0]
            else:
                threshold = _percentile(data=reg_based, q=per_dic[-std_value])
                ind = np.where(reg_based >= threshold)[0]

            if return_ind:
                dic[key] = tuple(ind.tolist())
            else:
                dic[key] = tuple(val[ind].tolist())
        self.outlier_results = dic
        return self

    def add_outlier_distance(self, plus: Optional[bool] = True, std_value: int = 2, return_ind: bool = True):
        per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
        dic = {val: self.modeling_data.x_data[:, ind] for ind, val in enumerate(self.modeling_data.x_data_names)}
        for key, val in dic.items():
            arr = stack(val, self.modeling_data.y_data, False)
            cent_other = _cent(arr[:, 0], arr[:, 1])
            ran = range(0, self.modeling_data.len)
            x_y_other_centers = np.array([_dis(_cent(x_lst=[arr[i][0]], y_lst=[arr[i][1]]), cent_other) for i in ran])

            if plus:
                x_y_other_centers_std = _percentile(data=x_y_other_centers, q=per_dic[std_value])
                ind = np.where(x_y_other_centers <= x_y_other_centers_std)[0]
            else:
                x_y_other_centers_std = _percentile(data=x_y_other_centers, q=per_dic[-std_value])
                ind = np.where(x_y_other_centers >= x_y_other_centers_std)[0]

            if return_ind:
                dic[key] = tuple(ind.tolist())
            else:
                dic[key] = tuple(val[ind].tolist())
        self.outlier_results = dic
        return self

    def add_outlier_hist(self, plus: Optional[bool] = True, std_value: int = 2, return_ind: bool = True):
        per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
        dic = {val: self.modeling_data.x_data[:, ind] for ind, val in enumerate(self.modeling_data.x_data_names)}
        for key, val in dic.items():
            n, b = np.histogram(val, bins='sturges')
            if plus:
                qn = _percentile(data=val, q=per_dic[std_value])
                ind = np.where(n <= qn)[0]
                bin_edges = np.array([(b[i], b[i + 1]) for i in range(len(b) - 1)])[ind]
            else:
                qn = _percentile(data=val, q=per_dic[-std_value])
                ind = np.where(n >= qn)[0]
                bin_edges = np.array([(b[i], b[i + 1]) for i in range(len(b) - 1)])[ind]

            z_selected_ind = []
            for i, j in enumerate(val):
                for k, l in bin_edges:
                    if k >= j <= l:
                        z_selected_ind.append(i)
                        break

            # select = np.in1d(arr, arr[z_selected_ind])
            # return np.array([np.where(arr == i)[0][0] for i in arr[np.in1d(arr, arr[~select])]])
            if return_ind:
                dic[key] = tuple(z_selected_ind)
            else:
                dic[key] = tuple(val[z_selected_ind].tolist())
        self.outlier_results = dic
        return self

    def add_outlier_knn(self, plus: Optional[bool] = True, std_value: int = 2, return_ind: bool = True):
        per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
        dic = {val: self.modeling_data.x_data[:, ind] for ind, val in enumerate(self.modeling_data.x_data_names)}
        for key, val in dic.items():
            arr = stack(val, self.modeling_data.y_data, False)
            ran = range(0, self.modeling_data.len)
            test_centers = (_cent([arr[ind, 0]], [arr[ind, 1]]) for ind in ran)
            distances = [_dis(cent1=i, cent2=j) for i in test_centers for j in test_centers]
            if plus:
                threshold = _percentile(data=distances, q=per_dic[std_value])
                count_dic = {}
                for i, j in enumerate(arr):
                    temp = arr[i, :] <= threshold
                    count_dic[i] = _sum([1 for i in temp if i == True])
            else:
                threshold = _percentile(data=distances, q=per_dic[-std_value])
                count_dic = {}
                for i, j in enumerate(arr):
                    temp = arr[i, :] >= threshold
                    count_dic[i] = _sum([1 for i in temp if i == True])

            lst = []
            for i in _to_metatype(data=count_dic.values()):
                if isinstance(i, list):
                    for val1 in i:
                        lst.append(val1)
                else:
                    lst.append(i)
            if plus:
                val1 = _percentile(data=lst, q=per_dic[std_value])
                ind = np.where(np.array(lst) <= np.floor(val1))[0]
            else:
                val1 = _percentile(data=lst, q=per_dic[-std_value])
                ind = np.where(np.array(lst) >= np.floor(val1))[0]
            if return_ind:
                dic[key] = tuple(ind.tolist())
            else:
                dic[key] = tuple(val[ind].tolist())
        self.outlier_results = dic
        return self

    def add_outlier_cooks_distance(self, plus: bool = True, std_value: int = 2, return_ind: bool = True):
        per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
        dic = {val: self.modeling_data.x_data[:, ind] for ind, val in enumerate(self.modeling_data.x_data_names)}
        for key, val in dic.items():
            x = sm.add_constant(data=val)
            y = self.modeling_data.y_data
            model = sm.OLS(y, x).fit()
            np.set_printoptions(suppress=True)
            influence = model.get_influence()
            cooks = influence.cooks_distance
            if plus:
                val1 = _percentile(data=cooks[0], q=per_dic[std_value])
                ind = np.where(cooks[0] <= val1)[0]
            else:
                val1 = _percentile(data=cooks[0], q=per_dic[-std_value])
                ind = np.where(cooks[0] >= val1)[0]
            if return_ind:
                dic[key] = tuple(ind.tolist())
            else:
                dic[key] = tuple(val[ind].tolist())
        self.outlier_results = dic
        return self

    def get_outliers(self):
        ran = range(self.modeling_data.len)
        key_dic = dict(zip(self.modeling_data.x_data_names, range(len(self.modeling_data.x_data_names))))
        dic = self.outlier_results
        for key, val in dic.items():
            results = {i: True for i in val}
            temp = [i for i in ran if i not in results]
            dic[key] = tuple(self.modeling_data.x_data[:, key_dic[key]][temp].tolist())
        return dic



"""
Scatter plot class.

Usage:
 ./plot/scatter.py

Author:
 Peter Rigali - 2022-03-10
"""
from dataclasses import dataclass
from typing import List, Union, Tuple
import matplotlib.pyplot as plt
import pandas as pd
from scipy import stats
from pyjr.classes.data import Data
from pyjr.classes.preprocess_data import PreProcess
from pyjr.utils.tools import _to_metatype


@dataclass
class Scatter:
    """

    Class for plotting scatter plots.

    :param data: Input data.
    :type data: Either Data(pyjr class), PreProcess(pyjr class), list of the latter or a dict, pd.DataFrame.
    :param limit: Selection of rows to include.
    :type limit: list or tuple of int's.
    :param label_lst: Columns or names to focus on.
    :type label_lst: list or tuple of str's.
    :param color_lst: List or tuple of colors to use in plot.
    :type color_lst: List or tuple
    :param regression_line:  If included, requires a column str or List[str], default = None. *Optional*
    :type regression_line: List[str]
    :param regression_line_color: Color of regression line, default = 'red'. *Optional*
    :type regression_line_color: str
    :param regression_line_lineweight: Regression lineweight, default = 2.0. *Optional*
    :type regression_line_lineweight: float
    :param fig_size: default = (10, 7), *Optional*
    :type fig_size: tuple
    :param ylabel: Y axis label. *Optional*
    :type ylabel: str
    :param ylabel_color: Y axis label color, default = 'black'. *Optional*
    :type ylabel_color: str
    :param ylabel_size: Y label size, default = 'medium'. *Optional*
    :type ylabel_size: str
    :param xlabel: X axis label. *Optional*
    :type xlabel: str
    :param xlabel_color: X axis label color, default = 'black'. *Optional*
    :type xlabel_color: str
    :param xlabel_size: X label size, default = 'medium'. *Optional*
    :type xlabel_size: str
    :param title: Title of plt.
    :type title: str
    :param title_size: Size of title font. *Default is xx-large*
    :type title_size: str
    :param grid: If True will show grid, default = true. *Optional*
    :type grid: bool
    :param grid_alpha: Grid alpha, default = 0.75. *Optional*
    :type grid_alpha: float
    :param grid_dash_sequence: Grid dash sequence, default = (3, 3). *Optional*
    :type grid_dash_sequence: tuple
    :param grid_lineweight: Grid lineweight, default = 0.5. *Optional*
    :type grid_lineweight: float
    :param legend_fontsize: Legend fontsize, default = 'medium'. *Optional*
    :type legend_fontsize: str
    :param legend_transparency: Legend transparency, default = 0.75. *Optional*
    :type legend_transparency: float
    :param legend_location: legend location, default = 'lower right'. *Optional*
    :type legend_location: str
    :param compare_two: If given will return a scatter comparing two variables, default is None. *Optional*
    :type compare_two: List[str]
    :param y_limit: If given will limit the y axis.
    :type y_limit: List[float]
    :example: *None*
    :note:
        fonts can be: {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}
        location can be: {'best', 'upper right', 'upper left', 'lower left', 'lower right', 'right', 'center left',
                          'center right', 'lower center', 'upper center', 'center'}

    """
    __slots__ = "ax"

    def __init__(self,
                 data: Union[pd.DataFrame, Data, PreProcess, List[Union[Data, PreProcess]]],
                 limit: Union[List[int], Tuple[int]] = None,
                 label_lst: Union[List[str], Tuple[str]] = None,
                 color_lst: Union[List[str], Tuple[str]] = None,
                 regression_line: Union[List[str], Tuple[str]] = None,
                 regression_line_color: str = None,
                 regression_line_lineweight: float = 2.0,
                 fig_size: tuple = (10, 7),
                 ylabel: str = None,
                 ylabel_color: str = 'black',
                 ylabel_size: str = 'medium',
                 xlabel: str = None,
                 xlabel_color: str = 'black',
                 xlabel_size: str = 'medium',
                 title: str = 'Scatter Plot',
                 title_size: str = 'xx-large',
                 grid: bool = True,
                 grid_alpha: float = 0.75,
                 grid_dash_sequence: tuple = (3, 3),
                 grid_lineweight: float = 0.5,
                 legend_fontsize: str = 'medium',
                 legend_transparency: float = 0.75,
                 legend_location: str = 'lower right',
                 compare_two: Union[Tuple[str], bool] = None,
                 y_limit: Union[list, tuple] = None,
                 show: bool = False,
                 ):
        # Parse input data
        if isinstance(data, (Data, PreProcess)):
            if label_lst is None:
                label_lst = _to_metatype(data=data.name, dtype='list')
            data = data.dataframe()
        elif isinstance(data, pd.DataFrame):
            if label_lst is None:
                label_lst = _to_metatype(data=data.columns, dtype='list')
        elif isinstance(data, list):
            dic = {}
            for d in data:
                if isinstance(d.name, (list, tuple)):
                    for ind, val in enumerate(d.name):
                        dic[val.name] = val.data[:, ind]
                else:
                    dic[d.name] = d.data
            data = pd.DataFrame.from_dict(dic)
            label_lst = _to_metatype(data=data.columns, dtype='list')

        # Get colors
        if color_lst is None:
            if label_lst.__len__() <= 3:
                color_lst = ['tab:orange', 'tab:blue', 'tab:green'][:label_lst.__len__()]
            else:
                color_lst = [plt.get_cmap('viridis')(1. * i / label_lst.__len__()) for i in range(label_lst.__len__())]

        if regression_line_color is None:
            regression_line_color = 'r'

        # Start plot
        fig, ax = plt.subplots(figsize=fig_size)

        if limit:
            data = data[limit[0]:limit[1]]

        # Get compare two
        if compare_two is True:
            if _to_metatype(data=data.columns, dtype='tuple').__len__() >= 2:
                if ylabel is None:
                    ylabel = label_lst[1]
                if xlabel is None:
                    xlabel = label_lst[0]
                x_axis = data[label_lst[0]]
                label_lst = [label_lst[1]]
            else:
                raise AttributeError("If compare_two is True, input data requires two columns of data.")
        elif isinstance(compare_two, (list, tuple)):
            if compare_two.__len__() == 2:
                label_lst = [compare_two[1]]
                x_axis = data[compare_two[0]]
                if ylabel is None:
                    ylabel = compare_two[1]
                if xlabel is None:
                    xlabel = compare_two[0]
            else:
                raise AttributeError("If compare_two is a list, input data requires two columns of data.")
        else:
            x_axis = range(len(data))
            if ylabel is None:
                ylabel = 'Values'
            if xlabel is None:
                xlabel = 'Index'

        # Get plots
        count = 0
        for ind in label_lst:
            d = data[ind]
            ax.scatter(x=x_axis, y=d, color=color_lst[count], label=ind)
            if regression_line is not None and ind in regression_line:
                slope, intercept, r_value, p_value, std_err = stats.linregress(x_axis, d)
                ax.plot(x_axis, intercept + slope * x_axis, color=regression_line_color,
                        label=ind+'_ols_'+str(round(slope, 2)), linestyle='--', linewidth=regression_line_lineweight)
            count += 1
        ax.set_ylabel(ylabel, color=ylabel_color, fontsize=ylabel_size)
        ax.tick_params(axis='y', labelcolor=ylabel_color)
        ax.set_title(title, fontsize=title_size)

        # Add grid
        if grid:
            ax.grid(alpha=grid_alpha, linestyle=(0, grid_dash_sequence), linewidth=grid_lineweight)

        ax.set_xlabel(xlabel, color=xlabel_color, fontsize=xlabel_size)

        # Add legend
        if compare_two is None:
            ax.legend(fontsize=legend_fontsize, framealpha=legend_transparency, loc=legend_location, frameon=True)

        if y_limit:
            ax.set_ylim(bottom=y_limit[0], top=y_limit[1])

        self.ax = ax

        if show:
            plt.show()

    def __repr__(self):
        return 'Scatter Plot'



"""
Time Series class.

Usage:
 ./utils/timeseries.py

Author:
 Peter Rigali - 2022-03-19
"""
from dataclasses import dataclass
from typing import Union
import numpy as np
import math
from pyjr.classes.data import Data
from pyjr.classes.preprocess_data import PreProcess
from scipy import signal
from pyts.bag_of_words import BagOfWords, WordExtractor
from pyts.approximation import DiscreteFourierTransform, MultipleCoefficientBinning, PiecewiseAggregateApproximation
from pyts.approximation import SymbolicAggregateApproximation, SymbolicFourierApproximation
from pyts.decomposition import SingularSpectrumAnalysis
from pyts.transformation import BOSS, ShapeletTransform
from pyts.metrics import boss, dtw
from scipy.stats import ks_2samp, ttest_ind
from pyjr.utils.base import _min


@dataclass
class TimeSeries:

    __slots__ = ["cleanData", "data", "name"]

    def __init__(self, data: Union[Data, PreProcess]):
        self.cleanData = data
        self.data = data.data
        self.name = "TS_" + data.name

    def __repr__(self):
        return "TimeSeriesData"

    def add_fft(self):
        y, Pxx = signal.periodogram(self.data, fs=self.cleanData.len, window='hanning', scaling='spectrum')
        Pxx = np.argsort(np.abs(Pxx))[::-1][1:10]
        self.data = tuple([1 / y[i] for i in Pxx])
        return self

    def add_zero_crossing(self, rate: int = 100):
        sig = self.data
        indices = find((sig[1:] >= 0) & (sig[:-1] < 0))
        crossings = [i - sig[i] / (sig[i + 1] - sig[i]) for i in indices]
        self.data = np.mean(np.diff(crossings) / rate)
        return self

    def add_bag_of_words(self, window_size: Union[int, int] = 10, word_size: Union[int, float] = 4,
                         strategy: str = 'normal', n_bins: int = 4):
        # {'uniform': bins have identical widths, 'quantile': Same number of points., 'normal': Bin edges are from normal dist.}
        bow = BagOfWords(window_size=window_size, word_size=word_size, strategy=strategy, n_bins=n_bins)
        self.data = bow.transform(self.data)
        self.name = "TS_BagOfWords"
        return self

    def add_word_extractor(self, window_size: Union[int, float] = 0.1, window_step: Union[int, float] = 1):
        word = WordExtractor(window_size=window_size, window_step=window_step)
        self.data = word.transform(self.data)
        self.name = "TS_WordExtractor"
        return self

    def add_DFT(self, n_coef: Union[int, float] = None):
        transfomer = DiscreteFourierTransform(n_coefs=n_coef)
        self.data = transfomer.fit_transform(self.data)
        self.name = "TS_DiscreteFouruerTransform"
        return self

    def add_MCB(self, n_bins: int = 4, strategy: str = 'quantile'):
        # {'uniform': bins have identical widths, 'quantile': Same number of points., 'normal': Bin edges are from normal dist., 'entropy': Bin edges calc through information gain}
        transformer = MultipleCoefficientBinning(n_bins=n_bins, strategy=strategy)
        self.data = transformer.fit_transform(self.data)
        self.name = "TS_MultipleCoefficientBinning"
        return self

    def add_SAX(self, n_bins: int = 4, strategy: str = 'quantile'):
        # {'uniform': bins have identical widths, 'quantile': Same number of points., 'normal': Bin edges are from normal dist.}
        transformer = SymbolicAggregateApproximation(n_bins=n_bins, strategy=strategy)
        self.data = transformer.fit_transform(self.data)
        self.name = "TS_SymbolicAggregateApproximation"
        return self

    def add_PAA(self, window_size: Union[int, float] = 4):
        transformer = PiecewiseAggregateApproximation(window_size=window_size)
        self.data = transformer.fit_transform(self.data)
        self.name = "TS_PiecewiseAggregateApproximation"
        return self

    def add_SFA(self, n_bins: int = 4, strategy: str = 'quantile'):
        # {'uniform': bins have identical widths, 'quantile': Same number of points., 'normal': Bin edges are from normal dist., 'entropy': Bin edges calc through information gain}
        transformer = SymbolicFourierApproximation(n_bins=n_bins, strategy=strategy)
        self.data = transformer.fit_transform(self.data)
        self.name = "TS_SymbolicFourierApproximation"
        return self

    def add_SSA(self, window_size: Union[int, float] = 4):
        transformer = SingularSpectrumAnalysis(window_size=window_size)
        self.data = transformer.fit_transform(self.data)
        self.name = "TS_SingularSpectrumAnalysis"
        return self

    def add_BOSS(self, window_size: Union[int, int] = 10, word_size: Union[int, float] = 4,
                 strategy: str = 'normal', n_bins: int = 4):
        # {'uniform': bins have identical widths, 'quantile': Same number of points., 'normal': Bin edges are from normal dist. 'entropy'}
        bow = BOSS(window_size=window_size, word_size=word_size, strategy=strategy, n_bins=n_bins)
        self.data = bow.fit_transform(self.data)
        self.name = "TS_BOSS"
        return self

    def add_Shapelet(self, data, criterion: str = "mutual_info"):
        # "anova"
        bow = ShapeletTransform(criterion=criterion)
        self.data = bow.fit_transform(self.data, data.data)
        self.name = "TS_ShapeletTransform"
        return self

    def get_boss(self, data):
        return boss(x=self.data, y=data.data)

    def get_dtw(self, data, dist: str = "square", method: str = "classic"):
        # {"square", "absolute", "precomputed", "callable"}
        # {"classic", "sakoechiba", "itakura", "region", "multiscale", "fast"}
        # more options if .show_options()
        return dtw(x=self.data, y=data.data, dist=dist, method=method)

    # Distance
    def get_euclidean_distance(self, other) -> float:
        """Returns the distance between two lists or tuples."""
        if isinstance(self.data, (list, tuple)) and isinstance(other.data, (list, tuple)):
            return math.dist(self.data, other.data) / self.data.__len__()
        else:
            raise AttributeError("Both data and other data need to be a list or tuple.")

    # MAPE
    # can not handle zero in denominator
    def get_mape(self, other, min_value: float = 0.01) -> float:
        """Returns the MAPE between two lists or tuples."""
        if isinstance(self.data, (list, tuple)) and isinstance(other.data, (list, tuple)):
            if _min(data=self.data) == 0:
                actual = np.array([i if i != 0.0 else min_value for i in self.data])
            else:
                actual = self.data
            if _min(data=other.data) == 0:
                pred = np.array([i if i != 0.0 else min_value for i in other.data])
            else:
                pred = other.data
            return np.mean(np.abs((actual - pred) / actual)) * 100
        else:
            raise AttributeError("Both data and other data need to be a list or tuple.")

    # Auto-Correlation
    def get_auto_corr(self, other, num: int = 50) -> float:
        """Returns the correlation of auto-corr between two lists or tuples."""
        if isinstance(self.data, (list, tuple)) and isinstance(other.data, (list, tuple)):
            def acf(x, length):
                return [1] + [np.corrcoef(x[:-i], x[i:])[0,1] for i in range(1, length)]
            return np.corrcoef(acf(x=self.data, length=num), acf(x=other.data, length=num))[0, 1]
        else:
            raise AttributeError("Both data and other data need to be a list or tuple.")

    # Correlation
    def get_corr(self, other) -> float:
        """Returns the correlation of two lists or tuples."""
        if isinstance(self.data, (list, tuple)) and isinstance(other.data, (list, tuple)):
            return np.corrcoef(self.data, other.data)[0, 1]
        else:
            raise AttributeError("Both data and other data need to be a list or tuple.")

    # T Test to compare means
    def get_compare_means(self, other) -> float:
        """Returns the t-test, comparing means of two lists or tuples."""
        if isinstance(self.data, (list, tuple)) and isinstance(other.data, (list, tuple)):
            return ttest_ind(a=self.data, b=other.data).pvalue
        else:
            raise AttributeError("Both data and other data need to be a list or tuple.")

    # Kolmogorov-smirnov to see if from the same distribution
    def get_kol_smirnov(self, other) -> float:
        """Returns if teo lists or tuples come from the same distribution, as a pvalue."""
        if isinstance(self.data, (list, tuple)) and isinstance(other.data, (list, tuple)):
            return ks_2samp(data1=self.data, data2=other.data).pvalue
        else:
            raise AttributeError("Both data and other data need to be a list or tuple.")



"""
Regression class object.

Usage:
 ./classes/simple_regression.py

Author:
 Peter Rigali - 2022-03-19
"""
from dataclasses import dataclass
from typing import Union
import numpy as np
from statsmodels import regression
from pyjr.classes.data import Data
from pyjr.classes.preprocess_data import PreProcess
from pyjr.utils.tools import _add_constant


@dataclass
class Regression:
    """

    Calculate linear regression.

    :param x_data: Input x Data.
    :type x_data: Data or PreProcess
    :param y_data: Input y Data.
    :type y_data: Data or PreProcess
    :example:
    :note: This will return a Regression object with regression result information.

    """
    __slots__ = "results"

    def __init__(self, x_data: Union[Data, PreProcess], y_data: Union[Data, PreProcess]):
        if x_data.len != y_data.len:
            raise AttributeError('X and Y data are not the same length.')
        model = regression.linear_model.OLS(y_data.array(), _add_constant(x_data.data)).fit()
        self.results = {att: getattr(model, att) for att in dir(model) if '__' not in att and att[0] is not '_' and isinstance(getattr(model, att), (np.float_, np.ndarray, np.int_, np.str_, np.bool_))}

    def __repr__(self):
        return 'RegressionAnalysis'



"""
Histogram plot class.

Usage:
 ./plot/histogram.py

Author:
 Peter Rigali - 2022-03-10
"""
from dataclasses import dataclass
from typing import List, Union, Tuple
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from scipy.stats import norm
from pyjr.classes.data import Data
from pyjr.classes.preprocess_data import PreProcess
from pyjr.utils.tools import _to_metatype
from pyjr.utils.base import _mean, _std


@dataclass
class Histogram:
    """

    Class for plotting histograms.

    :param data: Input data.
    :type data: Either Data(pyjr class), PreProcess(pyjr class), list of the latter or a dict, pd.DataFrame.
    :param limit: Selection of rows to include.
    :type limit: list or tuple of int's.
    :param label_lst: Columns or names to focus on.
    :type label_lst: list or tuple of str's.
    :param color_lst: List or tuple of colors to use in plot.
    :type color_lst: List or tuple
    :param include_norm: Include norm. If included, requires a column str, default = None. *Optional*
    :type include_norm: str
    :param norm_color: Norm color, default = 'red'. *Optional*
    :type norm_color: str
    :param norm_lineweight: Norm lineweight, default = 1.0. *Optional*
    :type norm_lineweight: float
    :param norm_ylabel: Norm Y axis label. *Optional*
    :type norm_ylabel: str
    :param norm_legend_location: Location of norm legend, default = 'upper right'. *Optional*
    :type norm_legend_location: str
    :param fig_size: default = (10, 7), *Optional*
    :type fig_size: tuple
    :param bins: Way of calculating bins, default = 'sturges'. *Optional*
    :type bins: str
    :param hist_type: Type of histogram, default = 'bar'. *Optional*
    :type hist_type: str
    :param stacked: If True, will stack histograms, default = False. *Optional*
    :type stacked: bool
    :param ylabel: Y axis label. *Optional*
    :type ylabel: str
    :param ylabel_color: Y axis label color, default = 'black'. *Optional*
    :type ylabel_color: str
    :param ylabel_size: Y label size, default = 'medium'. *Optional*
    :type ylabel_size: str
    :param ytick_rotation:
    :type ytick_rotation: int = 0,
    :param xlabel: X axis label. *Optional*
    :type xlabel: str
    :param xlabel_color: X axis label color, default = 'black'. *Optional*
    :type xlabel_color: str
    :param xlabel_size: X label size, default = 'medium'. *Optional*
    :type xlabel_size: str
    :param xtick_rotation:
    :type xtick_rotation: int = 0,
    :param title: Title of plt.
    :type title: str
    :param title_size: Size of title font. *Default is xx-large*
    :type title_size: str
    :param grid: Whether to include a grid. *Default is True.*
    :type grid: bool
    :param grid_alpha: Transparency of grid lines. *Default is 0.75'*
    :type grid_alpha: float
    :param grid_lineweight: Width of gridlines. *Default is 0.5.*
    :type grid_lineweight: float
    :param grid_dash_sequence: Tuple of spaces and lines in grid. *Default is (1, 3)*
    :type grid_dash_sequence: tuple
    :param fig_size: Width and height of figure. *Default is (10, 7).*
    :type fig_size: tuple
    :param legend_fontsize: Legend fontsize, default = 'medium'. *Optional*
    :type legend_fontsize: str
    :param legend_transparency: Legend transparency, default = 0.75. *Optional*
    :type legend_transparency: float
    :param legend_location: legend location, default = 'lower right'. *Optional*
    :type legend_location: str
    :param show: Whether to print the plot. *Default is False.*
    :type show: bool
    :example: *None*
    :note:
        fonts can be: {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}
        location can be: {'best', 'upper right', 'upper left', 'lower left', 'lower right', 'right', 'center left',
                          'center right', 'lower center', 'upper center', 'center'}
    """
    __slots__ = ("ax", "ax1")

    def __init__(self,
                 data: Union[pd.DataFrame, Data, PreProcess, List[Union[Data, PreProcess]]],
                 color_lst: Union[List[str], Tuple[str]] = None,
                 label_lst: Union[List[str], Tuple[str]] = None,
                 limit: Union[List[int], Tuple[int]] = None,
                 include_norm: str = None,
                 norm_color: str = 'r',
                 norm_lineweight: float = 1.0,
                 norm_ylabel: str = None,
                 norm_legend_location: str = 'upper right',
                 fig_size: tuple = (10, 7),
                 bins: str = 'sturges',
                 hist_type: str = 'bar',
                 stacked: bool = False,
                 ylabel: str = None,
                 ylabel_color: str = 'black',
                 ylabel_size: str = 'medium',
                 ytick_rotation: int = 0,
                 xlabel: str = None,
                 xlabel_color: str = 'black',
                 xlabel_size: str = 'medium',
                 xtick_rotation: int = 0,
                 title: str = 'Histogram',
                 title_size: str = 'xx-large',
                 grid: bool = True,
                 grid_alpha: float = 0.75,
                 grid_dash_sequence: tuple = (3, 3),
                 grid_lineweight: float = 0.5,
                 legend_fontsize: str = 'medium',
                 legend_transparency: float = 0.75,
                 legend_location: str = 'lower right',
                 show: bool = False,
                 ):
        # Parse input data
        if isinstance(data, (Data, PreProcess)):
            if label_lst is None:
                label_lst = _to_metatype(data=data.name, dtype='list')
            data = data.dataframe()
        elif isinstance(data, pd.DataFrame):
            if label_lst is None:
                label_lst = _to_metatype(data=data.columns, dtype='list')
        elif isinstance(data, list):
            dic = {}
            for d in data:
                if isinstance(d.name, (list, tuple)):
                    for ind, val in enumerate(d.name):
                        dic[val.name] = val.data[:, ind]
                else:
                    dic[d.name] = d.data
            data = pd.DataFrame.from_dict(dic)
            label_lst = _to_metatype(data=data.columns, dtype='list')

        # Get colors
        if color_lst is None:
            if label_lst.__len__() <= 3:
                color_lst = ['tab:orange', 'tab:blue', 'tab:green'][:label_lst.__len__()]
            else:
                color_lst = [plt.get_cmap('viridis')(1. * i / label_lst.__len__()) for i in range(label_lst.__len__())]

        # Start plot
        fig, ax = plt.subplots(figsize=fig_size)

        if limit:
            data = data[limit[0]:limit[1]]

        # Get plots
        count = 0
        for ind in label_lst:
            ax.hist(data[ind], bins=bins, color=color_lst[count], label=ind, stacked=stacked, histtype=hist_type)
            count += 1
        ax.set_ylabel(ylabel, color=ylabel_color, fontsize=ylabel_size)
        ax.tick_params(axis='y', labelcolor=ylabel_color, rotation=ytick_rotation)
        ax.set_title(title, fontsize=title_size)

        # Add grid
        if grid:
            ax.grid(alpha=grid_alpha, linestyle=(0, grid_dash_sequence), linewidth=grid_lineweight)

        ax.set_xlabel(xlabel, color=xlabel_color, fontsize=xlabel_size)
        ax.tick_params(axis='x', labelcolor=ylabel_color, rotation=xtick_rotation)
        ax.legend(fontsize=legend_fontsize, framealpha=legend_transparency, loc=legend_location, frameon=True)

        # Plot normal curve
        ax1 = None
        if include_norm:
            d = _to_metatype(data=data[include_norm], dtype='list')
            _mu, _s = norm.fit(np.random.normal(_mean(data=d), _std(data=d), d.__len__()))
            xmin, xmax = plt.xlim()
            x = np.linspace(xmin, xmax, 100)
            ax1 = ax.twinx()
            ax1.plot(x, norm.pdf(x, _mu, _s), color=norm_color, linewidth=norm_lineweight, linestyle='--',
                     label="{} Fit Values: mu {:.2f} and std {:.2f}".format(include_norm, _mu, _s))
            ax1.set_ylabel(norm_ylabel, color=norm_color)
            ax1.tick_params(axis='y', labelcolor=norm_color)
            ax1.legend(fontsize=legend_fontsize, framealpha=legend_transparency, loc=norm_legend_location, frameon=True)

        self.ax = ax
        self.ax1 = ax1

        if show:
            plt.show()

    def __repr__(self):
        return 'Histogram Plot'



"""
Handling errors.

Usage:
 ./utils/tools.py

Author:
 Peter Rigali - 2022-03-19
"""
from typing import Union, List, Optional
import numpy as np
from collections.abc import KeysView, ValuesView

import pandas as pd
from pandas import Series, DataFrame


# Repeat functions
def _mean_(data: Union[list, tuple]) -> float:
    """
    Find the mean value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :return: Mean value.
    :rtype: float.
    :note: *None*
    """
    return sum(data) / data.__len__()


def _variance_(data: Union[list, tuple], ddof: int = 1) -> float:
    """
    Find the variance value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :param ddof: Desired Degrees of Freedom.
    :type ddof: int
    :return: Variance value.
    :rtype: float.
    :note: *None*
    """
    mu = _mean_(data=data)
    return sum(((x - mu) ** 2 for x in data)) / (data.__len__() - ddof)


def _std_(data: list, ddof: int = 1) -> float:
    """
    Find the Standard Deviation value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :param ddof: Desired Degrees of Freedom.
    :type ddof: int
    :return: Standard Deviation value.
    :rtype: float.
    :note: *None*
    """
    return _variance_(data=data, ddof=ddof) ** .5


def _percentile_(data: Union[list, tuple], q: float) -> float:
    """
    Find the percentile value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :param q: Percentile percent.
    :type q: float.
    :return: Percentile value.
    :note: *None*
    """
    data = _round_to(data=[item * 1000.0 for item in data], val=1)
    ind = _round_to(data=data.__len__() * q, val=1)
    data.sort()
    for item in data:
        if item >= data[ind]:
            return item / 1000.0


# Cleaning Functions.
def _empty(data) -> bool:
    """Checks if data is empty"""
    if data.__len__() == 0:
        return True
    else:
        return False


def _to_metatype(data, dtype: str = 'list') -> Union[list, tuple]:
    """Converts list-adjacent objects to a list or tuple"""
    if dtype not in {'list': True, 'tuple': True}:
        raise AttributeError("dtype input must be either {list, tuple}.")
    if isinstance(data, {'list': list, 'tuple': tuple}[dtype]):
        return data
    elif isinstance(data, Series):
        if dtype == 'list':
            return data.to_list()
        else:
            return tuple(data.to_list())
    elif isinstance(data, np.ndarray):
        if dtype == 'list':
            return data.tolist()
        else:
            return tuple(data.tolist())
    elif isinstance(data, (set, KeysView, ValuesView, pd.Index)):
        return {'list': list, 'tuple': tuple}[dtype](data)
    elif isinstance(data, (int, float, str, object, np.int_, np.float_, np.str, np.object)):
        if dtype == 'list':
            return [data]
        else:
            return tuple(data)
    else:
        raise AttributeError('Input data needs to have a type of {np.ndarray, pd.Series, list, set, int, float, str, object}')


def _to_type(value: Union[float, int, str, object], dtype: str = 'float') -> Union[float, int, str, object]:
    """Converts value to a set type"""
    if isinstance(value, {'float': float, 'int': int, 'str': str, 'object': object}[dtype]):
        return value
    else:
        return {'float': float, 'int': int, 'str': str, 'object': object}[dtype](value)


def _check_type(data: Union[list, tuple], dtype: str = 'float') -> tuple:
    """Checks type of values in a list"""
    return tuple([_to_type(value=val, dtype=dtype) for val in data])


def _check_na(value) -> bool:
    """
    Checks a value to see if Nan.

    :param value: Input value.
    :return: Returns True or False if the value is Nan.
    :rtype: bool.
    :note: *None*
    """
    if value == value and value is not None and value != np.inf and value != -np.inf:
        return False
    else:
        return True


def _remove_nan(data: Union[list, tuple]) -> list:
    """Remove Nan values from a list"""
    return [val for val in data if _check_na(val) is False]


def _round_to(data: Union[list, Series, np.ndarray, float, int], val: float,
              remainder: bool = False) -> Union[list, float]:
    """
    Rounds a value or list.

    :param data: Value or list.
    :param val: Place to round to.
    :param remainder: Whether to use remainder. If using floats, this should be true.
    # :param val_type: Desired value type.
    # :type val_type: str.
    :return: Returns a value or list of values.
    :note: *None*
    """
    if isinstance(data, (float, int)):
        if remainder is True:
            return round(_to_type(value=data, dtype='float') * val) / val
        else:
            return round(_to_type(value=data, dtype='float') / val) * val
    elif isinstance(data, (list, Series, np.ndarray)):
        data = (_to_type(value=i, dtype='float') for i in _to_metatype(data=data))
        if remainder is True:
            return [round(item * val) / val for item in data]
        else:
            return [round(item / val) * val for item in data]
    else:
        raise AttributeError('Value not one of the specified types.')


def _replacement_value(data: list, na_handling: str = 'median', std_value: int = 3, cap_zero: bool = True,
                       median_value: float = 0.023, ddof: int = 1) -> Union[float, None]:
    """
    Calculate desired replacement for Nan values.

    :param data: Input data.
    :type data: list.
    :param na_handling: Desired Nan value handling method. {zero, mu, std, median}
    :type na_handling: str.
    :param std_value: Desired Standard Deviation to use.
    :type std_value: int.
    :param cap_zero: Whether to cap the value at zero.
    :type cap_zero: bool.
    :param median_value: Desired percentile to use.
    :type median_value: float.
    :return: Replacement value.
    :note: If mean - 3 * std is less than 0, may confuse results.
    """
    if na_handling == 'zero':
        return 0.0
    elif na_handling == 'mu':
        return _mean_(data=_remove_nan(data=data))
    elif na_handling == 'std':
        data = _remove_nan(data=data)
        val = _mean_(data=data) - (_std_(data=data, ddof=ddof) * std_value)
        if cap_zero:
            if val > 0:
                return val
            else:
                return 0.0
        else:
            return val
    elif na_handling == 'median':
        return _percentile_(data=_remove_nan(data=data), q=median_value)
    elif na_handling == 'none':
        return None


def _replace_na(data: Union[list, tuple], replacement_value: float = None) -> Union[list, tuple]:
    """Replace Nan values with replacement value"""
    if replacement_value is None:
        return _remove_nan(data=data)
    else:
        return tuple([val if _check_na(value=val) is False else replacement_value for val in data])


# def _prep(data, dtype: str = 'float', na_handling: str = 'median', std_value: int = 3, median_value: float = 0.023,
#           cap_zero: bool = True, ddof: int = 1):
#     """Cleans data"""
#     if _empty(data=data) is False:
#         data = _to_metatype(data=data)
#         na_value = _replacement_value(data=data, na_handling=na_handling, std_value=std_value,
#                                       median_value=median_value, cap_zero=cap_zero, ddof=ddof)
#         return _check_type(data=_replace_na(data=data, replacement_value=na_value), dtype=dtype)
#     else:
#         return None


def _prep(data, meta_type: str = 'tuple', dtype: str = 'float', na_handling: str = 'zero', std_value: int = 3,
          median_value: float = 0.023, cap_zero: bool = True, ddof: int = 1):
    """Clean data"""
    # Check Empty
    if _empty(data=data):
        raise AttributeError("Data entered is empty.")
    # Convert to list
    data = _to_metatype(data=data, dtype='list')
    # Check type, check na, replace na
    na = None
    for ind, val in enumerate(data):
        if _check_na(value=val):
            if na is None:
                na = _replacement_value(data=data, na_handling=na_handling)
            val = na
        data[ind] = _to_type(value=val, dtype=dtype)
    # Convert to metadata
    data = _to_metatype(data=data, dtype=meta_type)
    return data


# Non-cleaning related functions
def _add_constant(data: Union[list, tuple, np.ndarray]) -> np.ndarray:
    """Add a column of ones to a list, tuple or np.ndarray"""
    if isinstance(data, (tuple, list)):
        arr = np.ones((data.__len__(), 2))
    elif isinstance(data, np.ndarray):
        arr = np.ones((data.shape[0], 2))
    arr[:, 1] = data
    return arr


def _unique_values(data: Union[list, tuple], count: False):
    """
    Finds unique values from a list.

    :param data: Input data.
    :type data: list or tuple.
    :return: Returns either a list or dict.
    :note: *None*
    """
    if count:
        unique = set(data)
        return {i: data.count(i) for i in unique}
    else:
        return tuple(set(data))


def _search_dic_values(dic: dict, item: Union[str, int, float]) -> Union[str, float, int]:
    """

    Searches a dict using the values.

    :param dic: Input data.
    :type dic: dict
    :param item: Search item.
    :type item: str, float or int
    :return: Key value connected to the value.
    :rtype: str, float or int
    :example: *None*
    :note: *None*

    """
    return list(dic.keys())[list(dic.values()).index(item)]


# Model_data
def _check_names(name: str, name_list: Union[list, tuple]) -> bool:
    name_dic = {name: True for name in name_list}
    if name not in name_dic:
        return True
    else:
        raise AttributeError("{} already included in names list".format(name))


def _check_len(len1: int, len2: int) -> bool:
    if len1 == len2:
        return True
    else:
        raise AttributeError("(len1: {} ,len2: {} ) Lengths are not the same.".format(len1, len2))


def _add_column(arr1: np.ndarray, arr2: np.ndarray) -> np.ndarray:
    new_arr = np.ones((arr1.shape[0], arr1.shape[1] + 1))
    for i in range(arr1.shape[1]):
        new_arr[:, i] = arr1[:, i]
    new_arr[:, new_arr.shape[1] - 1] = arr2[:, 0]
    return new_arr


# FeaturePerformance
def _cent(x_lst: List[float], y_lst: List[float]) -> List[float]:
    """

    Calculate Centroid from x and y value(s).

    :param x_lst: A list of values.
    :type x_lst: List[float]
    :param y_lst: A list of values.
    :type y_lst: List[float]
    :returns: A list of x and y values representing the centriod of two lists.
    :rtype: List[float]
    :example: *None*
    :note: *None*

    """
    return [np.sum(x_lst) / len(x_lst), np.sum(y_lst) / len(y_lst)]


def _dis(cent1: List[float], cent2: List[float]) -> float:
    """

    Calculate Distance between two centroids.

    :param cent1: An x, y coordinate representing a centroid.
    :type cent1: List[float]
    :param cent2: An x, y coordinate representing a centroid.
    :type cent2: List[float]
    :returns: A distance measurement.
    :rtype: float
    :example: *None*
    :note: *None*

    """
    return round(np.sqrt((cent1[0] - cent2[0]) ** 2 + (cent1[1] - cent2[1]) ** 2), 4)


def stack(x_arr: np.ndarray, y_arr: np.ndarray, multi: Optional[bool] = False) -> np.ndarray:
    """

    Stacks x_arr and y_arr.

    :param x_arr: An array to stack.
    :type x_arr: np.ndarray
    :param y_arr: An array to stack.
    :type y_arr: np.ndarray
    :param multi: If True, will stack based on multiple x_arr columns, default is False. *Optional*
    :type multi: bool
    :return: Array with a x column and a y column
    :rtype: np.ndarray
    :example: *None*
    :note: *None*

    """
    lst = []
    if multi:
        for i in range((x_arr.shape[1])):
            lst.append(np.vstack([x_arr[:, i].ravel(), y_arr[:, i].ravel()]).T)
        return np.array(lst)
    else:
        lst = np.vstack([x_arr.ravel(), y_arr.ravel()]).T
    return np.where(np.isnan(lst), 0, lst)



"""
Component functions.

Usage:
 ./utils/base.py

Author:
 Peter Rigali - 2022-03-10
"""
from typing import Union
from pyjr.utils.tools import _unique_values, _search_dic_values, _round_to, _to_metatype


# Internal Math Functions
def _max(data: Union[list, tuple]) -> float:
    """
    Find the max value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :return: Maximum value.
    rtype: float.
    :note: *None*
    """
    if data.__len__() > 1:
        return max(data)
    return data[0]


def _min(data: Union[list, tuple]):
    """
    Find the min value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :return: Minimum value.
    :rtype: float.
    :note: *None*
    """
    if data.__len__() > 1:
        return min(data)
    return data[0]


def _range(data: Union[list, tuple]) -> float:
    """
    Find the max to min range value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :return: Range value.
    :rtype: float.
    :note: *None*
    """
    if data.__len__() > 1:
        return _max(data=data) - _min(data=data)
    return 0.0


def _mean(data: Union[list, tuple]) -> float:
    """
    Find the mean value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :return: Mean value.
    :rtype: float.
    :note: *None*
    """
    return sum(data) / data.__len__()


def _variance(data: Union[list, tuple], ddof: int = 1) -> float:
    """
    Find the variance value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :param ddof: Desired Degrees of Freedom.
    :type ddof: int
    :return: Variance value.
    :rtype: float.
    :note: *None*
    """
    mu = _mean(data=data)
    return sum(((x - mu) ** 2 for x in data)) / (data.__len__() - ddof)


def _std(data: Union[list, tuple], ddof: int = 1) -> float:
    """
    Find the Standard Deviation value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :param ddof: Desired Degrees of Freedom.
    :type ddof: int
    :return: Standard Deviation value.
    :rtype: float.
    :note: *None*
    """
    return _variance(data=data, ddof=ddof) ** .5


def _sum(data: Union[list, tuple]) -> float:
    """
    Find the sum value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :return: Sum value.
    rtype: float.
    :note: *None*
    """
    if data.__len__() > 1:
        return sum(data)
    return data[0]


def _median(data: Union[list, tuple]) -> float:
    """
    Find the median value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :return: Mean value.
    :rtype: float.
    :note: *None*
    """
    sorted_lst, lst_len = sorted(data), data.__len__()
    index = (lst_len - 1) // 2
    if lst_len % 2:
        return sorted_lst[index]
    else:
        return _mean(data=[sorted_lst[index]] + [sorted_lst[index + 1]])


def _mode(data: Union[list, tuple]) -> float:
    """
    Find the mode value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :return: Mode value.
    :rtype: float.
    :note: *None*
    """
    count_dic = _unique_values(data=data, count=True)
    count_dic_values = _to_metatype(data=count_dic.values())
    dic_max = _max(count_dic_values)
    lst = []
    for i in count_dic_values:
        val = _search_dic_values(dic=count_dic, item=dic_max)
        lst.append((val, i))
        # del count_dic[val]
        count_dic_values = _to_metatype(data=count_dic.values())

    first_val, second_val = lst[0][0], lst[0][1]
    equal_lst = [i[0] for i in lst if second_val == i[1]]
    if equal_lst.__len__() == 1:
        return first_val
    elif equal_lst.__len__() % 2 == 0:
        return _mean(data=equal_lst)
    else:
        return _median(data=equal_lst)


def _skew(data: Union[list, tuple]) -> float:
    """
    Find the skew value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :return: Skew value.
    :rtype: float.
    :note: *None*
    """
    mu, stdn, length = _mean(data=data), _std(data=data, ddof=1) ** 3, data.__len__()
    if stdn == 0:
        stdn = mu / 2.0
    return (((_sum(data=[i - mu for i in data]) ** 3) / length) / stdn) * ((length * (length - 1)) ** .5) / (length - 2)


def _kurtosis(data: Union[list, tuple]) -> float:
    """
    Find the kurtosis value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :return: Kurtosis value.
    :rtype: float.
    :note: *None*
    """
    mu, stdn = _mean(data=data), _std(data=data, ddof=1) ** 4
    if stdn == 0:
        stdn = mu / 2
    return (((_sum(data=[i - mu for i in data])**4) / data.__len__()) / stdn) - 3


def _percentile(data: Union[list, tuple], q: float) -> float:
    """
    Find the percentile value of a list.

    :param data: Input data.
    :type data: list or tuple.
    :param q: Percentile percent.
    :type q: float.
    :return: Percentile value.
    rtype: float.
    :note: *None*
    """
    data = _round_to(data=[item * 1000.0 for item in data], val=1)
    ind = _round_to(data=data.__len__() * q, val=1)
    data.sort()
    for item in data:
        if item >= data[ind]:
            return item / 1000.0


def _percentiles(data: Union[list, tuple], q_lst: Union[list, tuple] = (0.159, 0.841)):
    """
    Calculate various percentiles for a list.

    :param data: Input data.
    :type data: list or tuple.
    :param q_lst: Desired percentile percents.
    :type q_lst: List of floats.
    :return: A group of stats.
    :note: *None*
    """
    return (_percentile(data=data, q=q) for q in q_lst)



"""
Classifiers class.

Usage:
 ./models/classification.py

Author:
 Peter Rigali - 2022-03-19
"""
from dataclasses import dataclass
from typing import Union
from pyjr.classes.model_data import ModelingData
from sklearn.linear_model import RidgeClassifier, SGDClassifier, LogisticRegression
from sklearn import svm
from sklearn.neighbors import KNeighborsClassifier, NearestCentroid
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.naive_bayes import GaussianNB
from pyts.classification import TSBF, SAXVSM, BOSSVS
from pyts.multivariate.classification import MultivariateClassifier
from sklearn.metrics import accuracy_score, roc_auc_score, confusion_matrix, f1_score, precision_score, recall_score


@dataclass
class Classifier:

    __slots__ = ["data", "pred", "coef", "intercept", "scores", "name"]

    def __init__(self, data: ModelingData):
        self.data = data
        self.pred = None
        self.coef = None
        self.intercept = None
        self.scores = None
        self.name = None

    def __repr__(self):
        return "Classifier"

    def add_ridge(self):
        cls = RidgeClassifier().fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = cls.coef_
        self.intercept = cls.intercept_
        self.pred = cls.predict(X=self.data.x_test)
        self.name = "CLS_RidgeClassifier"
        return self

    def add_svc(self, kernel: str = 'rbf', gamma: str = 'scale'):
        """Benefits from standardizing"""
        # {‘linear’, ‘poly’, ‘rbf’, ‘sigmoid’, ‘precomputed’}
        # {'scale', 'auto'}
        cls = svm.SVC(kernel=kernel, gamma=gamma).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = cls.coef_
        self.intercept = cls.intercept_
        self.pred = cls.predict(X=self.data.x_test)
        self.name = "CLS_SVC"
        return self

    def add_nusvc(self, kernel: str = 'rbf', gamma: str = 'scale'):
        """Benefits from standardizing"""
        # {‘linear’, ‘poly’, ‘rbf’, ‘sigmoid’, ‘precomputed’}
        # {'scale', 'auto'}
        cls = svm.NuSVC(kernel=kernel, gamma=gamma).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = cls.coef_
        self.intercept = cls.intercept_
        self.pred = cls.predict(X=self.data.x_test)
        self.name = "CLS_NuSVC"
        return self

    def add_svm(self):
        """Benefits from standardizing"""
        reg = svm.LinearSVC().fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "CLS_LinearSVC"
        return self

    def add_sgdclass(self, loss: str = 'hinge'):
        # The possible options are ‘hinge’, ‘log’, ‘modified_huber’, ‘squared_hinge’, ‘perceptron’, or a
        # regression loss: ‘squared_error’, ‘huber’, ‘epsilon_insensitive’, or ‘squared_epsilon_insensitive’.
        """Benefits from standardizing"""
        reg = SGDClassifier(loss=loss).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "CLS_SGDClassifier"
        return self

    def add_knnclass(self, num: int = 5, weights: str = 'uniform'):
        reg = KNeighborsClassifier(n_neighbors=num, weights=weights).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "CLS_KNeighborsClassifier"
        return self

    def add_knncentroid(self):
        reg = NearestCentroid().fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "CLS_NearestCentroid"
        return self

    def add_gaussian(self):
        # from sklearn.gaussian_process.kernels import RBF
        reg = GaussianProcessClassifier(random_state=0).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "CLS_GaussianProcessClassifier"
        return self

    def add_tree(self):
        reg = DecisionTreeClassifier(random_state=0).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "CLS_DecisionTreeClassifier"
        return self

    def add_ada(self):
        reg = AdaBoostClassifier(random_state=0).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "CLS_AdaBoostClassifier"
        return self

    def add_forest(self):
        reg = RandomForestClassifier(random_state=0).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "CLS_RandomForestClassifier"
        return self

    def add_mlp(self, activation: str = 'relu', solver: str = 'adam', learning_rate: str = 'constant'):
        # {‘identity’, ‘logistic’, ‘tanh’, ‘relu’}
        # {‘lbfgs’, ‘sgd’, ‘adam’}
        reg = MLPClassifier(random_state=0, activation=activation, solver=solver, learning_rate=learning_rate)
        reg.fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercepts_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "CLS_MLPClassifier"
        return self

    def add_nb(self):
        reg = GaussianNB().fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "CLS_GaussianNB"
        return self

    def add_log(self):
        reg = LogisticRegression(random_state=0).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "CLS_LogisticRegression"
        return self

    # TimeSeries
    def add_tsbf(self, random_state: int = 0):
        clf = TSBF(random_state=random_state).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = clf.feature_importance_
        self.pred = clf.pedict(X=self.data.x_test)
        self.intercept = None
        self.name = "CLS_TSBF"
        return self

    def add_szxvsm(self, window_size: Union[int, float] = 0.5, word_size: Union[int, float] = 0.5, n_bins: int = 4,
                   stategy: str = "normal"):
        # {'uniform': bins have identical widths, 'quantile': Same number of points.,
        # 'normal': Bin edges are from normal dist.}
        clf = SAXVSM(window_size=window_size, word_size=word_size, n_bins=n_bins,
                     strategy=stategy).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = None
        self.pred = clf.pedict(X=self.data.x_test)
        self.intercept = None
        self.name = "CLS_SAXVSM"
        return self

    def add_bossvs(self, window_size: int = 4, word_size: int = 4, n_bins: int = 4):
        clf = BOSSVS(window_size=window_size, word_size=word_size, n_bins=n_bins).fit(X=self.data.x_train,
                                                                                      y=self.data.y_train)
        self.coef = None
        self.pred = clf.pedict(X=self.data.x_test)
        self.intercept = None
        self.name = "CLS_BOSSVS"
        return self

    def add_mc(self, estimator=BOSSVS()):
        # {estimator object or list of}
        clf = MultivariateClassifier(estimator=estimator).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = None
        self.pred = clf.pedict(X=self.data.x_test)
        self.intercept = None
        self.name = "CLS_MultivariateClassifier"
        return self

    # Metrics
    def add_scores(self):
        self.scores = {"accuracy": accuracy_score(y_true=self.data.y_test, y_pred=self.pred),
                       "f1": f1_score(y_true=self.data.y_test, y_pred=self.pred),
                       "recall": recall_score(y_true=self.data.y_test, y_pred=self.pred),
                       "precision": precision_score(y_true=self.data.y_test, y_pred=self.pred),
                       "roc_auc": roc_auc_score(y_true=self.data.y_test, y_score=self.pred),
                       "conf_matrix": confusion_matrix(y_true=self.data.y_test, y_pred=self.pred)}
        return self



"""
Stand-alone functions.

Usage:
 ./utils/simple.py

Author:
 Peter Rigali - 2022-03-19
"""
from dataclasses import dataclass
import statsmodels.api as sm
from typing import Union, Optional
from pandas import Series
import numpy as np
from pyjr.utils.base import _percentile, _mean, _variance, _sum
from pyjr.utils.tools import _prep, _unique_values, _to_type, _replace_na, _replacement_value, _dis, _cent, stack, _to_metatype
from pyjr.classes.data import Data
from pyjr.classes.preprocess_data import PreProcess


def _clean(data, dtype, na):
    if na is None:
        na = _replacement_value(data=data, na_handling='zero')
    else:
        na = _replacement_value(data=data, na_handling=na)
    data = _replace_na(data=data, replacement_value=na)
    if dtype is None:
        dtype = np.min_scalar_type(data).type
    else:
        type_tup = (np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64, np.float16,
                    np.float32, np.float64, np.float_, np.str_, np.int_)
        test_dic = {i.__name__: True for i in type_tup}
        if dtype in test_dic:
            dtype = {i.__name__: i for i in type_tup}[dtype]
            if dtype in {np.float_: True, np.int_: True, np.bool_: True}:
                data = dtype(data).tolist()
            else:
                data = np.array(data).astype(dtype).tolist()
        else:
            raise AttributeError("Must use numpy dtypes")
    return data


def _one_hot_encode(func):
    def wrapper(*args, **kwargs):
        data, dtype, na = func(*args, **kwargs)
        data = _clean(data, dtype, na)
        unique = _unique_values(data=data, count=False)
        arr = np.zeros((len(data), len(unique)))
        for ind in range(len(unique)):
            arr[:, ind] = [1.0 if str(ind) == str(val) else 0.0 for val in data]
        return arr
    return wrapper


@_one_hot_encode
def oneHotEncode(data: list, dtype: str = "str_", na: str = None):
    """One hot encode a list of data"""
    return (data, dtype, na)


def calc_gini(data: Union[list, np.ndarray, Series],
              na_handling: str = 'none',
              dtype: str = 'float',
              std_value: int = 3,
              median_value: float = 0.023,
              cap_zero: bool = True,
              ddof: int = 1) -> Union[float, int]:
    """

    Calculate the Gini Coef for a list.

    :param data: Input data.
    :type data: list, np.ndarray, or pd.Series
    :return: Gini value.
    :rtype: float
    :example:
        >>> lst = [4.3, 5.6]
        >>> calc_gini(data=lst, val=4, remainder=True) # 0.05445544554455435
    :note: The larger the gini coef, the more consolidated the chips on the table are to one person.
    """
    new_data = _prep(data=data, dtype=dtype, na_handling=na_handling, std_value=std_value,
                     median_value=median_value, cap_zero=cap_zero, ddof=ddof)
    sorted_list = sorted(new_data)
    height, area = 0.0, 0.0
    for value in sorted_list:
        height += value
        area += height - value / 2.0
    fair_area = height * len(new_data) / 2.0
    return _to_type(value=(fair_area - area) / fair_area, dtype=dtype)


def outlier_std(data, plus: bool = True, std_value: int = 2, return_ind: bool = False) -> np.ndarray:
    """

    Calculate Outliers using a simple std value.

    :param arr: An Array to get data from. *Optional*
    :type arr: np.ndarray
    :param data: A DataFrame to get data from. *Optional*
    :type data: pd.DataFrame
    :param y_column: A target column. *Optional*
    :type y_column: str
    :param _std: A std threshold, default is 3. *Optional*
    :type _std: int
    :param plus: If True, will grab all values above the threshold, default is True. *Optional*
    :type plus: bool
    :return: An array of indexes.
    :rtype: np.ndarray
    :example: *None*
    :note: If **arr** not passed, data and respective column names are required.

    """
    per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
    new_data = np.array(data.data)
    if data.min >= 0:
        if plus:
            ind = np.where(new_data <= _percentile(data=data.data, q=per_dic[std_value]))[0]
        else:
            ind = np.where(new_data >= _percentile(data=data.data, q=per_dic[-std_value]))[0]
    else:
        if plus:
            ind = np.where(new_data <= data.mean + data.std * std_value)[0]
        else:
            ind = np.where(new_data >= data.mean - data.std * std_value)[0]

    if return_ind:
        return ind
    else:
        return new_data[ind]


def outlier_var(data: Data, plus: Optional[bool] = True, std_value: int = 2,
                return_ind: bool = False) -> np.ndarray:
    """

    Calculate Outliers using a simple var value.

    :param arr: An Array to get data from. *Optional*
    :type arr: np.ndarray
    :param data: A DataFrame to get data from. *Optional*
    :type data: pd.DataFrame
    :param y_column: A target column. *Optional*
    :type y_column: str
    :param per: A percent threshold, default is 0.95. *Optional*
    :type per: float
    :param plus: If True, will grab all values above the threshold. *Optional*
    :type plus: bool, default is True
    :return: An array of indexes.
    :rtype: np.ndarray
    :example: *None*
    :note: If **arr** not passed, data and respective column names are required.

    """
    per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
    lst = data.data.copy()
    temp_var = _variance(data=lst, ddof=data.inputs['ddof'])
    dev_based = np.array([temp_var - _variance(np.delete(lst, i), ddof=data.inputs['ddof']) for i, j in enumerate(lst)])

    if plus:
        q = _percentile(data=lst, q=per_dic[std_value])
        ind = np.where(dev_based <= q)[0]
    else:
        q = _percentile(data=lst, q=per_dic[-std_value])
        ind = np.where(dev_based >= q)[0]

    if return_ind:
        return ind
    else:
        return np.array(lst)[ind]


def outlier_regression(x_data: Data, y_data: Data, plus: Optional[bool] = True, std_value: Optional[int] = 2,
                       return_ind: bool = False) -> np.ndarray:
    """

    Calculate Outliers using regression.

    :param arr: An Array to get data from. *Optional*
    :type arr: np.ndarray
    :param data: A DataFrame to get data from. *Optional*
    :type data: pd.DataFrame
    :param x_column: A column for x variables. *Optional*
    :type x_column: str
    :param y_column: A column for y variables. *Optional*
    :type y_column: str
    :param std_value: A std threshold, default is 3. *Optional*
    :type std_value: int
    :param plus: If True, will grab all values above the threshold, default is True. *Optional*
    :type plus: bool
    :return: An array of indexes.
    :rtype: np.ndarray
    :example: *None*
    :note: If **arr** not passed, data and respective column names are required.

    """
    per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
    arr = stack(np.array(x_data.data), np.array(y_data.data), False)
    ran = np.array(range(x_data.len))
    mu_y = np.zeros(len(arr) - 1)
    line_ys = []
    for i, j in enumerate(arr):
        xx, yy = np.delete(arr[:, 0], i), np.delete(arr[:, 1], i)
        w1 = (np.cov(xx, yy, ddof=1) / _variance(xx, ddof=1))[0, 1]
        new_y = w1 * ran[:-1] + (-1 * _mean(xx) * w1 + _mean(yy))
        mu_y = (mu_y + new_y) / 2
        line_ys.append(new_y)

    reg_based = np.array([np.mean(np.square(mu_y - j)) for i, j in enumerate(line_ys)])
    if plus:
        threshold = _percentile(data=reg_based, q=per_dic[std_value])
        ind = np.where(reg_based <= threshold)[0]
    else:
        threshold = _percentile(data=reg_based, q=per_dic[-std_value])
        ind = np.where(reg_based >= threshold)[0]

    if return_ind:
        return ind
    else:
        return arr[ind]


def outlier_distance(x_data: Data, y_data: Data, plus: Optional[bool] = True, std_value: int = 2,
                     return_ind: bool = False) -> np.ndarray:
    """

    Calculate Outliers using distance measurements.

    :param arr: An Array to get data from. *Optional*
    :type arr: np.ndarray
    :param: data: A DataFrame to get data from. *Optional*
    :type data: pd.DataFrame
    :param x_column: A column for x variables. *Optional*
    :type x_column: str
    :param y_column: A column for y variables. *Optional*
    :type y_column: str
    :param std_value: A std threshold, default is 3. *Optional*
    :type std_value: int
    :param plus: If True, will grab all values above the threshold, default is True. *Optional*
    :type plus: bool
    :return: An array of indexes.
    :rtype: np.ndarray
    :example: *None*
    :note: If **arr** not passed, data and respective column names are required.

    """
    per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
    arr = stack(np.array(x_data.data), np.array(y_data.data), False)
    cent_other = _cent(arr[:, 0], arr[:, 1])
    ran = range(0, x_data.len)
    x_y_other_centers = np.array([_dis(_cent(x_lst=[arr[i][0]], y_lst=[arr[i][1]]), cent_other) for i in ran])

    if plus:
        x_y_other_centers_std = _percentile(data=x_y_other_centers, q=per_dic[std_value])
        ind = np.where(x_y_other_centers <= x_y_other_centers_std)[0]
    else:
        x_y_other_centers_std = _percentile(data=x_y_other_centers, q=per_dic[-std_value])
        ind = np.where(x_y_other_centers >= x_y_other_centers_std)[0]

    if return_ind:
        return ind
    else:
        return arr[ind]


def outlier_hist(data: Data, plus: Optional[bool] = True, std_value: int = 2, return_ind: bool = False) -> np.ndarray:
    """

    Calculate Outliers using Histogram.

    :param arr: An Array to get data from. *Optional*
    :type arr: np.ndarray
    :param: data: A DataFrame to get data from. *Optional*
    :type data: pd.DataFrame
    :param x_column: A column for x variables. *Optional*
    :type x_column: str
    :param per: A std threshold, default is 3. *Optional*
    :type per: float
    :param plus: If True, will grab all values above the threshold, default is 0.75. *Optional*
    :type plus: bool
    :return: An array of indexes.
    :rtype: np.ndarray
    :example: *None*
    :note: If **arr** not passed, data and respective column names are required.

    """
    per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
    arr = np.array(data.data)
    n, b = np.histogram(arr, bins='sturges')

    if plus:
        qn = _percentile(data=data.data, q=per_dic[std_value])
        ind = np.where(n <= qn)[0]
        bin_edges = np.array([(b[i], b[i + 1]) for i in range(len(b) - 1)])[ind]
    else:
        qn = _percentile(data=data.data, q=per_dic[-std_value])
        ind = np.where(n >= qn)[0]
        bin_edges = np.array([(b[i], b[i + 1]) for i in range(len(b) - 1)])[ind]

    z_selected_ind = []
    for i, j in enumerate(arr):
        for k, l in bin_edges:
            if k >= j <= l:
                z_selected_ind.append(i)
                break

    # select = np.in1d(arr, arr[z_selected_ind])
    # return np.array([np.where(arr == i)[0][0] for i in arr[np.in1d(arr, arr[~select])]])
    if return_ind:
        return z_selected_ind
    else:
        return arr[z_selected_ind]


def outlier_knn(x_data: Data, y_data: Data, plus: Optional[bool] = True, std_value: int = 2,
                return_ind: bool = False) -> np.ndarray:
    """

    Calculate Outliers using KNN.

    :param arr: An Array to get data from. *Optional*
    :type arr: np.ndarray
    :param: data: A DataFrame to get data from. *Optional*
    :type data: pd.DataFrame
    :param x_column: A column for x variables. *Optional*
    :type x_column: str
    :param y_column: A column for y variables. *Optional*
    :type y_column: str
    :param std_value: A std threshold, default is 3. *Optional*
    :type std_value: int
    :param plus: If True, will grab all values above the threshold, default is True. *Optional*
    :type plus: bool
    :return: An array of indexes.
    :rtype: np.ndarray
    :example: *None*
    :note: If **arr** not passed, data and respective column names are required.

    """
    per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
    arr = stack(np.array(x_data.data), np.array(y_data.data), False)
    ran = range(0, x_data.len)
    test_centers = (_cent([arr[ind, 0]], [arr[ind, 1]]) for ind in ran)
    distances = [_dis(cent1=i, cent2=j) for i in test_centers for j in test_centers]

    if plus:
        threshold = _percentile(data=distances, q=per_dic[std_value])
        count_dic = {}
        for i, j in enumerate(arr):
            temp = arr[i, :] <= threshold
            count_dic[i] = _sum([1 for i in temp if i == True])
    else:
        threshold = _percentile(data=distances, q=per_dic[-std_value])
        count_dic = {}
        for i, j in enumerate(arr):
            temp = arr[i, :] >= threshold
            count_dic[i] = _sum([1 for i in temp if i == True])

    lst = []
    for val in _to_metatype(data=count_dic.values()):
        if isinstance(val, list):
            for val1 in val:
                lst.append(val1)
        else:
            lst.append(val)

    if plus:
        val1 = _percentile(data=lst, q=per_dic[std_value])
        ind = np.where(np.array(lst) <= np.floor(val1))[0]
    else:
        val1 = _percentile(data=lst, q=per_dic[-std_value])
        ind = np.where(np.array(lst) >= np.floor(val1))[0]

    if return_ind:
        return ind
    else:
        return arr[ind]


def outlier_cooks_distance(x_data: Data, y_data: Data, plus: bool = True, std_value: int = 2, return_ind: bool = False):
    per_dic = {-3: 0.001, -2: 0.023, -1: 0.159, 0: 0.50, 1: 0.841, 2: 0.977, 3: 0.999}
    x = sm.add_constant(data=x_data.data)
    y = y_data.data
    model = sm.OLS(y, x).fit()
    np.set_printoptions(suppress=True)
    influence = model.get_influence()
    cooks = influence.cooks_distance

    if plus:
        val1 = _percentile(data=cooks[0], q=per_dic[std_value])
        ind = np.where(cooks[0] <= val1)[0]
    else:
        val1 = _percentile(data=cooks[0], q=per_dic[-std_value])
        ind = np.where(cooks[0] >= val1)[0]

    if return_ind:
        return ind
    else:
        return np.array(x_data.data)[ind]



"""
Regressor class.

Usage:
 ./models/regression.py

Author:
 Peter Rigali - 2022-03-19
"""
from dataclasses import dataclass
from pyjr.classes.model_data import ModelingData
from sklearn.linear_model import LinearRegression, Ridge, Lasso, ElasticNet, LassoLars, BayesianRidge, ARDRegression
from sklearn.linear_model import SGDRegressor, QuantileRegressor, HuberRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import AdaBoostRegressor, RandomForestRegressor
from sklearn.neural_network import MLPRegressor
from sklearn.metrics import explained_variance_score, max_error, mean_absolute_error, mean_squared_error
from sklearn.metrics import mean_squared_log_error, median_absolute_error, r2_score, mean_absolute_percentage_error


@dataclass
class Regressor:

    __slots__ = ["data", "pred", "coef", "intercept", "scores", "name"]

    def __init__(self, data: ModelingData):
        self.data = data
        self.pred = None
        self.coef = None
        self.intercept = None
        self.scores = None
        self.name = None

    def __repr__(self):
        return "Regressor"

    def add_linear(self):
        reg = LinearRegression().fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_LinearRegression"
        return self

    def add_ridge(self, alpha: float = 1.0):
        reg = Ridge(alpha=alpha).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_Ridge"
        return self

    def add_lasso(self, alpha: float = 1.0):
        reg = Lasso(alpha=alpha).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_Lasso"
        return self

    def add_lassolars(self, alpha: float = 1.0):
        reg = LassoLars(alpha=alpha).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_LassoLars"
        return self

    def add_elasticnet(self, alpha: float = 1.0):
        reg = ElasticNet(alpha=alpha, random_state=0).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_ElasticNet"
        return self

    def add_bayesridge(self):
        reg = BayesianRidge().fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_BayesianRidge"
        return self

    def add_ard(self):
        reg = ARDRegression().fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_ARDRegression"
        return self

    def add_sgdregress(self):
        """Benefits from standardizing"""
        reg = SGDRegressor().fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_SGDRegressor"
        return self

    def add_quantile(self, q: float = 0.5):
        reg = QuantileRegressor(quantile=q).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_QuantileRegressor"
        return self

    def add_knn(self, num: int = 5):
        reg = KNeighborsRegressor(n_neighbors=num).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_KNeighborsRegressor"
        return self

    def add_gaussian(self):
        # from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel
        # kernel = DotProduct() + WhiteKernel()
        reg = GaussianProcessRegressor(random_state=0).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_GaussianProcessRegressor"
        return self

    def add_tree(self):
        reg = DecisionTreeRegressor(random_state=0).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_DecisionTreeRegressor"
        return self

    def add_ada(self):
        reg = AdaBoostRegressor(random_state=0).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_AdaBoostRegressor"
        return self

    def add_forest(self):
        reg = RandomForestRegressor(random_state=0).fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_RandomForestRegressor"
        return self

    def add_mlp(self, activation: str = 'relu', solver: str = 'adam', learning_rate: str = 'constant'):
        # {‘identity’, ‘logistic’, ‘tanh’, ‘relu’}
        # {‘lbfgs’, ‘sgd’, ‘adam’}
        reg = MLPRegressor(random_state=0, activation=activation, solver=solver, learning_rate=learning_rate)
        reg.fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercepts_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_MLPRegressor"
        return self

    def add_huber(self):
        reg = HuberRegressor().fit(X=self.data.x_train, y=self.data.y_train)
        self.coef = reg.coef_
        self.intercept = reg.intercept_
        self.pred = reg.predict(X=self.data.x_test)
        self.name = "REG_HuberRegressor"
        return self

    def add_scores(self):
        # root mean square error prefered
        self.scores = {"explianed variance": explained_variance_score(y_true=self.data.y_test, y_pred=self.pred),
                       "max error": max_error(y_true=self.data.y_test, y_pred=self.pred),
                       "mean abs error": mean_absolute_error(y_true=self.data.y_test, y_pred=self.pred),
                       "mean squared error": mean_squared_error(y_true=self.data.y_test, y_pred=self.pred),
                       "mean squared log error": mean_squared_log_error(y_true=self.data.y_test, y_pred=self.pred),
                       "median abs error": median_absolute_error(y_true=self.data.y_test, y_pred=self.pred),
                       "r2": r2_score(y_true=self.data.y_test, y_pred=self.pred),
                       "mean abs per error": mean_absolute_percentage_error(y_true=self.data.y_test, y_pred=self.pred)}
        return self



"""
Table plot class.

Usage:
 ./plot/table.py

Author:
 Peter Rigali - 2022-03-10
"""
from dataclasses import dataclass
from typing import List, Union, Tuple
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import six
from pyjr.classes.data import Data
from pyjr.classes.preprocess_data import PreProcess
from pyjr.utils.tools import _to_metatype
from pyjr.utils.base import _min, _max


@dataclass
class Table:
    """

    Class for plotting tables.

    :param data: Input data.
    :type data: Either Data(pyjr class), PreProcess(pyjr class), list of the latter or a dict, pd.DataFrame.
    :param label_lst: Columns or names to focus on.
    :type label_lst: list or tuple of str's.
    :param fig_size: default = (10, 10), *Optional*
    :type fig_size: tuple
    :param font_size: Font size inside cells, default = 'medium'. *Optional*
    :type font_size: str
    :param font_color: Color of text inside cells, default is 'black'. *Optional*
    :type font_color: str
    :param col_widths: Width of columns, default = 0.30. *Optional*
    :type col_widths: float
    :param row_colors: Color of rows. *Optional*
    :type row_colors: str
    :param header_colors: Header of table color. *Optional*
    :type header_colors: str
    :param edge_color: Color of cell edges, default = 'w'. *Optional*
    :type edge_color: str
    :param sequential_cells: If True will color ever other row. *Default is False.*
    :type sequential_cells: bool
    :param color_map: Color map used in cells, default = 'Greens'. *Optional*
    :type color_map: str
    :example: *None*
    :note:
        fonts can be: {'xx-small', 'x-small', 'small', 'medium', 'large', 'x-large', 'xx-large'}
        location can be: {'best', 'upper right', 'upper left', 'lower left', 'lower right', 'right', 'center left',
                          'center right', 'lower center', 'upper center', 'center'}

    """
    __slots__ = "ax"

    def __init__(self,
                 data: Union[pd.DataFrame, Data, PreProcess, List[Union[Data, PreProcess]]],
                 label_lst: Union[List[str], Tuple[str]] = None,
                 limit: Union[List[int], Tuple[int]] = None,
                 fig_size: tuple = (10, 10),
                 font_size: str = 'medium',
                 col_widths: float = 0.30,
                 row_colors: str = None,
                 header_colors: str = None,
                 edge_color: str = 'w',
                 sequential_cells: bool = False,
                 color_map: str = 'Greens',
                 font_color: str = 'black',
                 show: bool = False,
                 ):
        # Parse input data
        if isinstance(data, (Data, PreProcess)):
            data = data.dataframe()
        elif isinstance(data, pd.DataFrame):
            pass
        elif isinstance(data, list):
            dic = {}
            for d in data:
                if isinstance(d.name, (list, tuple)):
                    for ind, val in enumerate(d.name):
                        dic[val.name] = val.data[:, ind]
                else:
                    dic[d.name] = d.data
            data = pd.DataFrame.from_dict(dic)
        data['index'] = _to_metatype(data=data.index, dtype='list')

        if limit:
            data = data[limit[0]:limit[1]]

        if row_colors is None:
            row_colors = ['#f1f1f2', 'w']
        if type(row_colors) is str:
            row_colors = [row_colors, 'w']
        if header_colors is None:
            header_colors = ['tab:blue', 'w']
        if type(header_colors) is str:
            header_colors = [header_colors, 'w']

        if label_lst is None:
            lst = _to_metatype(data=data.columns, dtype='list')
            lst.remove('index')
            label_lst = ['index'] + lst

        col_widths = [col_widths] * label_lst.__len__()
        colours = None
        if sequential_cells:
            color_lst = []
            for col in label_lst:
                if type(data[col].iloc[0]) != str and col != 'index':
                    d = _to_metatype(data=data[col], dtype='list')
                    _norm = plt.Normalize(_min(data=d) - 1, _max(data=d) + 1)
                    temp = plt.get_cmap(color_map)(_norm(data[col]))
                elif type(data[col].iloc[0]) == str and col != 'index':
                    temp = [(1.0, 1.0, 1.0, 1.0), (0.945, 0.945, 0.949, 1.0)] * data.__len__()
                else:
                    temp = [(0.121, 0.466, 0.705, 0.15), (0.121, 0.466, 0.705, 0.30)] * data.__len__()
                temp_lst = []
                for i in range(data.__len__()):
                    temp_lst.append(tuple(temp[i]))
                color_lst.append(temp_lst)
            colours = np.array(pd.DataFrame(color_lst).T)

        # Start plot
        fig, ax = plt.subplots(figsize=fig_size)
        table = ax.table(cellText=data.values, colLabels=label_lst, colWidths=col_widths, loc='center',
                         cellLoc='center', cellColours=colours)
        table.set_fontsize(font_size)
        for k, cell in six.iteritems(table._cells):
            r, c = k
            cell.set_edgecolor(edge_color)
            if r == 0:
                cell.set_text_props(weight='bold', color=header_colors[1])
                cell.set_facecolor(header_colors[0])
            else:
                if sequential_cells:
                    cell.set_facecolor(row_colors[r % len(row_colors)])
                if c != 0 and c != 'index':
                    cell.set_text_props(color=font_color)
        ax.axis('tight')
        ax.axis('off')
        fig.tight_layout()

        self.ax = ax

        if show:
            plt.show()

    def __repr__(self):
        return 'Table Plot'
